Namespace LinqToVisualTree
    ''' <summary>
    ''' Adapts a DependencyObject to provide methods required for generate
    ''' a Linq To Tree API
    ''' </summary>
    Public Class VisualTreeAdapter
        Implements ILinqTree(Of DependencyObject)

        Private _item As DependencyObject

        Public Sub New(ByVal item As DependencyObject)
            _item = item
        End Sub

        Public Function Children() As IEnumerable(Of DependencyObject) Implements ILinqTree(Of DependencyObject).Children
            Dim childrenCount As Integer = VisualTreeHelper.GetChildrenCount(_item)
            Dim returnValue As New List(Of DependencyObject)
            For i  = 0 To childrenCount - 1
                returnValue.Add(VisualTreeHelper.GetChild(_item, i))
            Next i
            Return returnValue
        End Function

        Public ReadOnly Property Parent As DependencyObject Implements ILinqTree(Of DependencyObject).Parent
            Get
                Return VisualTreeHelper.GetParent(_item)
            End Get
        End Property
    End Class
End Namespace

Namespace LinqToVisualTree
    ''' <summary>
    ''' Defines an interface that must be implemented to generate the LinqToTree methods
    ''' </summary>
    ''' <typeparam name="T"></typeparam>
    Public Interface ILinqTree(Of T)
        Function Children() As IEnumerable(Of T)
        ReadOnly Property Parent As T
    End Interface

    Public Module TreeExtensions
        ''' <summary>
        ''' Returns a collection of descendant elements.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function Descendants(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Dim adapter As ILinqTree(Of DependencyObject) = New VisualTreeAdapter(item)
            Dim returnValue As New List(Of DependencyObject)
            Dim returnValueChild  As New List(Of DependencyObject)
            For Each child In adapter.Children
                returnValueChild.Add(child)
                For Each grandChild In returnValue
                    returnValue.Add(grandChild)
                Next grandChild
            Next child
            Return returnValueChild
        End Function

        ''' <summary>
        ''' Returns a collection containing this element and all descendant elements.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function DescendantsAndSelf(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Dim returnValue As New List(Of DependencyObject)

            returnValue.Add(item)
            Dim returnValueChild As New List(Of DependencyObject)

            For Each child In returnValue
                returnValueChild.Add(child)
            Next child
            Return returnValueChild
        End Function

        ''' <summary>
        ''' Returns a collection of ancestor elements.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function Ancestors(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Dim adapter As ILinqTree(Of DependencyObject) = New VisualTreeAdapter(item)
            Dim returnValue As New List(Of DependencyObject)
            Dim parent = adapter.Parent
            Do While parent IsNot Nothing
                returnValue.Add(parent)
                adapter = New VisualTreeAdapter(parent)
                parent = adapter.Parent
            Loop
            Return returnValue
        End Function

        ''' <summary>
        ''' Returns a collection containing this element and all ancestor elements.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function AncestorsAndSelf(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Dim returnValue As New List(Of DependencyObject)
            returnValue.Add(item)
            Dim returnValueAncestor As New List(Of DependencyObject)
            For Each ancestor In returnValue
                returnValueAncestor.Add(ancestor)
            Next ancestor
            Return returnValueAncestor
        End Function

        ''' <summary>
        ''' Returns a collection of child elements.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function Elements(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Dim returnValue As New List(Of DependencyObject)
            Dim adapter As ILinqTree(Of DependencyObject) = New VisualTreeAdapter(item)
            For Each child In adapter.Children
                returnValue.Add(child)
            Next child
            Return returnValue
        End Function

        ''' <summary>
        ''' Returns a collection of the sibling elements before this node, in document order.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function ElementsBeforeSelf(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Dim returnValue As New List(Of DependencyObject)
            If item.Ancestors().FirstOrDefault Is Nothing Then
                Return returnValue
            End If
            For Each child In item.Ancestors().First.Elements()
                If child.Equals(item) Then
                    Exit For
                End If
                returnValue.Add(child)
            Next child
            Return returnValue
        End Function

        ''' <summary>
        ''' Returns a collection of the after elements after this node, in document order.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function ElementsAfterSelf(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Dim returnValue As New List(Of DependencyObject)
            If item.Ancestors().FirstOrDefault Is Nothing Then
                Return returnValue
            End If
            Dim afterSelf As Boolean = False
            For Each child In item.Ancestors().First.Elements()
                If afterSelf Then
                    returnValue.Add(child)
                End If

                If child.Equals(item) Then
                    afterSelf = True
                End If
            Next child
            Return returnValue
        End Function

        ''' <summary>
        ''' Returns a collection containing this element and all child elements.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function ElementsAndSelf(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Dim returnValue As New List(Of DependencyObject)
            returnValue.Add(item)
            Dim returnValueChild As New List(Of DependencyObject)
            For Each child In returnValue
                returnValueChild.Add(child)
            Next child
            Return returnValueChild
        End Function

        ''' <summary>
        ''' Returns a collection of descendant elements which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension()> _
        Public Function Descendants(Of T)(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Return item.Descendants.Where(Function(i) TypeOf i Is T).Cast(Of DependencyObject)()
        End Function

        ''' <summary>
        ''' Returns a collection of the sibling elements before this node, in document order
        ''' which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function ElementsBeforeSelf(Of T)(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Return item.ElementsBeforeSelf.Where(Function(i) TypeOf i Is T).Cast(Of DependencyObject)()
        End Function

        ''' <summary>
        ''' Returns a collection of the after elements after this node, in document order
        ''' which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function ElementsAfterSelf(Of T)(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Return item.ElementsAfterSelf.Where(Function(i) TypeOf i Is T).Cast(Of DependencyObject)()
        End Function

        ''' <summary>
        ''' Returns a collection containing this element and all descendant elements
        ''' which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function DescendantsAndSelf(Of T)(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Return item.DescendantsAndSelf.Where(Function(i) TypeOf i Is T).Cast(Of DependencyObject)()
        End Function

        ''' <summary>
        ''' Returns a collection of ancestor elements which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function Ancestors(Of T)(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Return item.Ancestors.Where(Function(i) TypeOf i Is T).Cast(Of DependencyObject)()
        End Function

        ''' <summary>
        ''' Returns a collection containing this element and all ancestor elements
        ''' which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function AncestorsAndSelf(Of T)(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Return item.AncestorsAndSelf.Where(Function(i) TypeOf i Is T).Cast(Of DependencyObject)()
        End Function

        ''' <summary>
        ''' Returns a collection of child elements which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function Elements(Of T)(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Return item.Elements.Where(Function(i) TypeOf i Is T).Cast(Of DependencyObject)()
        End Function

        ''' <summary>
        ''' Returns a collection containing this element and all child elements.
        ''' which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function ElementsAndSelf(Of T)(ByVal item As DependencyObject) As IEnumerable(Of DependencyObject)
            Return item.ElementsAndSelf.Where(Function(i) TypeOf i Is T).Cast(Of DependencyObject)()
        End Function
    End Module

    Public Module EnumerableTreeExtensions
        ''' <summary>
        ''' Applies the given function to each of the items in the supplied
        ''' IEnumerable.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Private Function DrillDown(ByVal items As IEnumerable(Of DependencyObject), ByVal [function] As Func(Of DependencyObject, IEnumerable(Of DependencyObject))) As IEnumerable(Of DependencyObject)
            Dim returnValue As New List(Of DependencyObject)
            For Each item In items
                For Each itemChild In [function](item)
                    returnValue.Add(itemChild)
                Next itemChild
            Next item
            Return returnValue
        End Function

        ''' <summary>
        ''' Applies the given function to each of the items in the supplied
        ''' IEnumerable, which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension>
        Public Function DrillDown(Of T As DependencyObject)(ByVal items As IEnumerable(Of DependencyObject), ByVal [function] As Func(Of DependencyObject, IEnumerable(Of DependencyObject))) As IEnumerable(Of DependencyObject)
            Dim returnValue As New List(Of DependencyObject)
            For Each item In items
                For Each itemChild In [function](item)
                    If TypeOf itemChild Is T Then
                        returnValue.Add(CType(itemChild, T))
                    End If
                Next itemChild
            Next item
            Return returnValue
        End Function

        ''' <summary>
        ''' Returns a collection of descendant elements.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function Descendants(ByVal items As IEnumerable(Of DependencyObject)) As IEnumerable(Of DependencyObject)
            Return items.DrillDown(Function(i) i.Descendants())
        End Function

        ''' <summary>
        ''' Returns a collection containing this element and all descendant elements.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function DescendantsAndSelf(ByVal items As IEnumerable(Of DependencyObject)) As IEnumerable(Of DependencyObject)
            Return items.DrillDown(Function(i) i.DescendantsAndSelf())
        End Function

        ''' <summary>
        ''' Returns a collection of ancestor elements.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function Ancestors(ByVal items As IEnumerable(Of DependencyObject)) As IEnumerable(Of DependencyObject)
            Return items.DrillDown(Function(i) i.Ancestors())
        End Function

        ''' <summary>
        ''' Returns a collection containing this element and all ancestor elements.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function AncestorsAndSelf(ByVal items As IEnumerable(Of DependencyObject)) As IEnumerable(Of DependencyObject)
            Return items.DrillDown(Function(i) i.AncestorsAndSelf())
        End Function

        ''' <summary>
        ''' Returns a collection of child elements.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function Elements(ByVal items As IEnumerable(Of DependencyObject)) As IEnumerable(Of DependencyObject)
            Return items.DrillDown(Function(i) i.Elements())
        End Function

        ''' <summary>
        ''' Returns a collection containing this element and all child elements.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function ElementsAndSelf(ByVal items As IEnumerable(Of DependencyObject)) As IEnumerable(Of DependencyObject)
            Return items.DrillDown(Function(i) i.ElementsAndSelf())
        End Function

        ''' <summary>
        ''' Returns a collection of descendant elements which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function Descendants(Of T As DependencyObject)(ByVal items As IEnumerable(Of DependencyObject)) As IEnumerable(Of DependencyObject)
            Return items.DrillDown(Of T)(Function(i) i.Descendants())
        End Function

        ''' <summary>
        ''' Returns a collection containing this element and all descendant elements.
        ''' which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function DescendantsAndSelf(Of T As DependencyObject)(ByVal items As IEnumerable(Of DependencyObject)) As IEnumerable(Of DependencyObject)
            Return items.DrillDown(Of T)(Function(i) i.DescendantsAndSelf())
        End Function

        ''' <summary>
        ''' Returns a collection of ancestor elements which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function Ancestors(Of T As DependencyObject)(ByVal items As IEnumerable(Of DependencyObject)) As IEnumerable(Of DependencyObject)
            Return items.DrillDown(Of T)(Function(i) i.Ancestors())
        End Function

        ''' <summary>
        ''' Returns a collection containing this element and all ancestor elements.
        ''' which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function AncestorsAndSelf(Of T As DependencyObject)(ByVal items As IEnumerable(Of DependencyObject)) As IEnumerable(Of DependencyObject)
            Return items.DrillDown(Of T)(Function(i) i.AncestorsAndSelf())
        End Function

        ''' <summary>
        ''' Returns a collection of child elements which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function Elements(Of T As DependencyObject)(ByVal items As IEnumerable(Of DependencyObject)) As IEnumerable(Of DependencyObject)
            Return items.DrillDown(Of T)(Function(i) i.Elements())
        End Function

        ''' <summary>
        ''' Returns a collection containing this element and all child elements.
        ''' which match the given type.
        ''' </summary>
        <System.Runtime.CompilerServices.Extension> _
        Public Function ElementsAndSelf(Of T As DependencyObject)(ByVal items As IEnumerable(Of DependencyObject)) As IEnumerable(Of DependencyObject)
            Return items.DrillDown(Of T)(Function(i) i.ElementsAndSelf())
        End Function
    End Module
End Namespace
